# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
# Copyright (C) 2024 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Click command-line interface for database management."""

import click
from flask.cli import with_appcontext
from sqlalchemy_utils.functions import create_database, database_exists, drop_database

from .proxies import current_db
from .utils import create_alembic_version_table, drop_alembic_version_table


def abort_if_false(ctx, param, value):
    """Abort command is value is False."""
    if not value:
        ctx.abort()


def render_url(url):
    """Render the URL for CLI output."""
    return url.render_as_string(hide_password=True)


#
# Database commands
#
@click.group(chain=True)
def db():
    """Database commands."""


@db.command()
@click.option("-v", "--verbose", is_flag=True, default=False)
@with_appcontext
def create(verbose):
    """Create tables."""
    click.secho("Creating all tables!", fg="yellow", bold=True)
    with click.progressbar(current_db.metadata.sorted_tables) as bar:
        for table in bar:
            if verbose:
                click.echo(" Creating table {0}".format(table))
            table.create(bind=current_db.engine, checkfirst=True)
    create_alembic_version_table()
    click.secho("Created all tables!", fg="green")


@db.command()
@click.option("-v", "--verbose", is_flag=True, default=False)
@click.option(
    "--yes-i-know",
    is_flag=True,
    callback=abort_if_false,
    expose_value=False,
    prompt="Do you know that you are going to drop the db?",
)
@with_appcontext
def drop(verbose):
    """Drop tables."""
    click.secho("Dropping all tables!", fg="red", bold=True)
    with click.progressbar(reversed(current_db.metadata.sorted_tables)) as bar:
        for table in bar:
            if verbose:
                click.echo(" Dropping table {0}".format(table))
            table.drop(bind=current_db.engine, checkfirst=True)
        drop_alembic_version_table()
    click.secho("Dropped all tables!", fg="green")


@db.command()
@with_appcontext
def init():
    """Create database."""
    displayed_database = render_url(current_db.engine.url)
    click.secho(f"Creating database {displayed_database}", fg="green")
    database_url = current_db.engine.url.render_as_string(hide_password=False)
    if not database_exists(database_url):
        create_database(database_url)


@db.command()
@click.option(
    "--yes-i-know",
    is_flag=True,
    callback=abort_if_false,
    expose_value=False,
    prompt="Do you know that you are going to destroy the db?",
)
@with_appcontext
def destroy():
    """Drop database."""
    displayed_database = render_url(current_db.engine.url)
    click.secho(f"Destroying database {displayed_database}", fg="red", bold=True)
    plain_url = current_db.engine.url.render_as_string(hide_password=False)
    if current_db.engine.name == "sqlite":
        try:
            drop_database(plain_url)
        except FileNotFoundError:
            click.secho("Sqlite database has not been initialised", fg="red", bold=True)
    else:
        drop_database(plain_url)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Shared database object for Invenio."""

from flask_sqlalchemy import SQLAlchemy as FlaskSQLAlchemy
from sqlalchemy import MetaData, event, util
from sqlalchemy.engine import Engine
from sqlalchemy.sql import text
from werkzeug.local import LocalProxy

NAMING_CONVENTION = util.immutabledict(
    {
        "ix": "ix_%(column_0_label)s",
        "uq": "uq_%(table_name)s_%(column_0_name)s",
        "ck": "ck_%(table_name)s_%(constraint_name)s",
        "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
        "pk": "pk_%(table_name)s",
    }
)
"""Configuration for constraint naming conventions."""

metadata = MetaData(naming_convention=NAMING_CONVENTION)
"""Default database metadata object holding associated schema constructs."""


class SQLAlchemy(FlaskSQLAlchemy):
    """Implement or overide extension methods."""

    def apply_driver_hacks(self, app, sa_url, options):
        """Call before engine creation."""
        # Don't forget to apply hacks defined on parent object.
        super(SQLAlchemy, self).apply_driver_hacks(app, sa_url, options)

        if sa_url.drivername == "sqlite":
            connect_args = options.setdefault("connect_args", {})

            if "isolation_level" not in connect_args:
                # disable pysqlite's emitting of the BEGIN statement entirely.
                # also stops it from emitting COMMIT before any DDL.
                connect_args["isolation_level"] = None

            if not event.contains(Engine, "connect", do_sqlite_connect):
                event.listen(Engine, "connect", do_sqlite_connect)
            if not event.contains(Engine, "begin", do_sqlite_begin):
                event.listen(Engine, "begin", do_sqlite_begin)

            from sqlite3 import register_adapter

            def adapt_proxy(proxy):
                """Get current object and try to adapt it again."""
                return proxy._get_current_object()

            register_adapter(LocalProxy, adapt_proxy)

        elif sa_url.drivername == "postgresql+psycopg2":  # pragma: no cover
            from psycopg2.extensions import adapt, register_adapter

            def adapt_proxy(proxy):
                """Get current object and try to adapt it again."""
                return adapt(proxy._get_current_object())

            register_adapter(LocalProxy, adapt_proxy)

        elif sa_url.drivername == "mysql+pymysql":  # pragma: no cover
            from pymysql import converters

            def escape_local_proxy(val, mapping):
                """Get current object and try to adapt it again."""
                return converters.escape_item(
                    val._get_current_object(),
                    self.engine.dialect.encoding,
                    mapping=mapping,
                )

            converters.conversions[LocalProxy] = escape_local_proxy
            converters.encoders[LocalProxy] = escape_local_proxy

        return sa_url, options


def do_sqlite_connect(dbapi_connection, connection_record):
    """Ensure SQLite checks foreign key constraints.

    For further details see "Foreign key support" sections on
    https://docs.sqlalchemy.org/en/latest/dialects/sqlite.html#foreign-key-support
    """
    # Enable foreign key constraint checking
    cursor = dbapi_connection.cursor()
    cursor.execute("PRAGMA foreign_keys=ON")
    cursor.close()


def do_sqlite_begin(dbapi_connection):
    """Ensure SQLite transaction are started properly.

    For further details see "Foreign key support" sections on
    https://docs.sqlalchemy.org/en/rel_1_0/dialects/sqlite.html#pysqlite-serializable # noqa
    """
    # emit our own BEGIN
    dbapi_connection.execute(text("BEGIN"))


db = SQLAlchemy(metadata=metadata)
"""Shared database instance using Flask-SQLAlchemy extension.

This object is initialized during initialization of ``InvenioDB``
extenstion that takes care about loading all entrypoints from key
``invenio_db.models``.
"""



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Force naming convention."""

import sqlalchemy as sa
from alembic import op, util
from sqlalchemy import inspect
from sqlalchemy.sql import text

# revision identifiers, used by Alembic.
revision = "35c1075e6360"
down_revision = "dbdbc1b19cf2"
branch_labels = ()
depends_on = None

NAMING_CONVENTION = sa.util.immutabledict(
    {
        "ix": "ix_%(column_0_label)s",
        "uq": "uq_%(table_name)s_%(column_0_name)s",
        "ck": "ck_%(table_name)s_%(constraint_name)s",
        "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
        "pk": "pk_%(table_name)s",
    }
)
"""Configuration for constraint naming conventions (v1.0.0b5)."""


def upgrade():
    """Upgrade database."""
    op.execute(text("COMMIT"))  # See https://bitbucket.org/zzzeek/alembic/issue/123
    ctx = op.get_context()
    metadata = ctx.opts["target_metadata"]
    metadata.naming_convention = NAMING_CONVENTION
    metadata.bind = ctx.connection.engine
    insp = inspect(ctx.connection.engine)

    for table_name in insp.get_table_names():
        if table_name not in metadata.tables:
            continue

        table = metadata.tables[table_name]

        ixs = {}
        uqs = {}
        fks = {}

        for ix in insp.get_indexes(table_name):
            ixs[tuple(ix["column_names"])] = ix
        for uq in insp.get_unique_constraints(table_name):
            uqs[tuple(uq["column_names"])] = uq
        for fk in insp.get_foreign_keys(table_name):
            fks[(tuple(fk["constrained_columns"]), fk["referred_table"])] = fk

        with op.batch_alter_table(
            table_name, naming_convention=NAMING_CONVENTION
        ) as batch_op:
            for c in list(table.constraints) + list(table.indexes):
                key = None
                if isinstance(c, sa.schema.ForeignKeyConstraint):
                    key = (tuple(c.column_keys), c.referred_table.name)
                    fk = fks.get(key)
                    if fk and c.name != fk["name"]:
                        batch_op.drop_constraint(fk["name"], type_="foreignkey")
                        batch_op.create_foreign_key(
                            op.f(c.name),
                            fk["referred_table"],
                            fk["constrained_columns"],
                            fk["referred_columns"],
                            **fk["options"]
                        )
                elif isinstance(c, sa.schema.UniqueConstraint):
                    key = tuple(c.columns.keys())
                    uq = uqs.get(key)
                    if uq and c.name != uq["name"]:
                        batch_op.drop_constraint(uq["name"], type_="unique")
                        batch_op.create_unique_constraint(
                            op.f(c.name), uq["column_names"]
                        )
                elif isinstance(c, sa.schema.CheckConstraint):
                    util.warn(
                        "Update {0.table.name} CHECK {0.name} " "manually".format(c)
                    )
                elif isinstance(c, sa.schema.Index):
                    key = tuple(c.columns.keys())
                    ix = ixs.get(key)
                    if ix and c.name != ix["name"]:
                        batch_op.drop_index(ix["name"])
                        batch_op.create_index(
                            op.f(c.name),
                            ix["column_names"],
                            unique=ix["unique"],
                        )
                elif (
                    isinstance(c, sa.schema.PrimaryKeyConstraint)
                    or c.name == "_unnamed_"
                ):
                    # NOTE we don't care about primary keys since they have
                    # specific syntax.
                    pass
                else:
                    raise RuntimeError("Missing {0!r}".format(c))


def downgrade():
    """Downgrade database."""
